"""animation module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

from math import acos
from design import *

class LinkageAnimation:

    def __init__(self, design, drv_index, linkage_type):
        # drv_index:     Driver index
        # fol_index:     Follower index
        # border:        Outline of rounded-rectangle coupler
        #                frame that grows and shrinks to
        #                encompass all circlepoints and
        #                sliderpoints
        self.design = design
        self.linkage_type = linkage_type
        if linkage_type == "double-crank":
            self.init_double_crank(design, drv_index)
        elif linkage_type == "double-slider":
            self.init_double_slider(design, drv_index)
        elif linkage_type == "slider-crank":
            self.init_slider_crank(design, drv_index)
        elif linkage_type == "crank-slider":
            self.init_crank_slider(design, drv_index)
        # Reference values, absolute coordinates, design position:
        # cpl_length:    Coupler link (driver to follower) length
        # phi_cpl_0:     Coupler direction
        # theta_0:       Position axis angle
        #
        # Reference values, coupler coordinates:
        # border_ctr_0:  Coupler border center 
        # border_tr_0:   Coupler border top-right
        #
        self.cpl_length = abs(self.cpl_0)
        self.phi_cpl_0 = direction(self.cpl_0)
        self.theta_0 = design.positions[0].angle
        design.create_coupler_frame()
        self.border_ctr_0 = design.coupler_border.center_uc
        self.border_tr_0 = design.coupler_border.top_right_uu

    def init_crank_slider(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # cpl_0:         Coupler link (driver to follower)
        #
        fol_index = drv_index = 0 # One each k and s
        self.init_with_driver_k(design, drv_index)
        self.init_with_follower_s(design, fol_index)
        self.cpl_0 = self.s_fol_0 - self.k_drv_0

    def init_double_crank(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # cpl_0:         Coupler link (driver to follower)
        #
        fol_index = 1 - drv_index
        self.init_with_driver_k(design, drv_index)
        self.init_with_follower_k(design, fol_index)
        self.cpl_0 = self.k_fol_0 - self.k_drv_0

    def init_double_slider(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # cpl_0:         Coupler link (driver to follower)
        #
        fol_index = 1 - drv_index
        self.init_with_driver_s(design, drv_index)
        self.init_with_follower_s(design, fol_index)
        self.cpl_0 = self.s_fol_0 - self.s_drv_0

    def init_slider_crank(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # cpl_0:         Coupler link (driver to follower)
        #
        fol_index = drv_index = 0 # One each s and k
        self.init_with_driver_s(design, drv_index)
        self.init_with_follower_k(design, fol_index)
        self.cpl_0 = self.k_fol_0 - self.s_drv_0

    def init_with_driver_k(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # k_drv_0:       Driver circlepoint center
        # m_drv:         Driver centerpoint center
        # drv_length:    Driver link length
        # to_axis_ctr:   Vector to position axis center from driver
        #
        self.k_drv_0 = design.absolute_coordinates(
            design.circlepoints[drv_index].center_uc, 0)
        self.m_drv= design.centerpoints[drv_index].center_uc
        self.drv_length = abs(self.k_drv_0 - self.m_drv)
        self.to_axis_ctr = (design.positions[0].center_uc
                            - self.k_drv_0)
        # Initialize variables
        self.k_drv = self.k_drv_0

    def init_with_driver_s(self, design, drv_index):
        # Reference values, absolute coordinates, design position:
        # s_drv_0:       Driver sliderpoint center
        # to_axis_ctr:   Vector to position axis center from driver
        #
        self.s_drv_0 = design.absolute_coordinates(
            design.sliderpoints[drv_index].center_uc, 0)
        self.to_axis_ctr = (design.positions[0].center_uc
                            - self.s_drv_0)
        # Initialize variables
        self.s_drv = self.s_drv_0

    def init_with_follower_k(self, design, fol_index):
        # Reference values, absolute coordinates, design position:
        # k_fol_0:       Follower circlepoint center
        # m_fol:         Follower centerpoint center
        # fol_length:    Follower link length
        self.k_fol_0 = design.absolute_coordinates(
            design.circlepoints[fol_index].center_uc, 0)
        self.m_fol = design.centerpoints[fol_index].center_uc
        self.fol_length= abs(self.k_fol_0 - self.m_fol)
        # Initialize variables
        self.k_fol = self.k_fol_0

    def init_with_follower_s(self, design, fol_index):
        # Reference values, absolute coordinates, design position:
        # s_fol_0:       Follower sliderpoint center
        # s_fol_angle    Follower sliderpoint direction
        self.s_fol_0 = design.absolute_coordinates(
            design.sliderpoints[fol_index].center_uc, 0)
        self.s_fol_angle = design.sliderpoints[fol_index].angle
        # Initialize variables
        self.s_fol = self.s_fol_0

    def append_sketches(self, figure_list, center_uc, angle):
        for sketch in self.design.coupler_sketches:
            # Be careful to copy the sketch object, not by reference 
            sketch_figure = Sketch(center_uc,
                                   sketch.geometry[0].points, "Anim",
                                   ANIMATION_COLOR)
            sketch_figure.angle = angle
            # figure_list is a reference, not a copy
            figure_list.append(sketch_figure)

    def draw(self, canvas, phi_drv, mouse_xy, pixels_per_unit,
             origin_cc):
        if self.linkage_type == "crank-slider":
            self.draw_crank_slider(
                canvas, phi_drv, pixels_per_unit, origin_cc)
        elif self.linkage_type == "double-crank":
            self.draw_double_crank(
                canvas, phi_drv, pixels_per_unit, origin_cc)
        elif self.linkage_type == "double-slider":
            self.draw_double_slider(
                canvas, phi_drv, mouse_xy, pixels_per_unit, origin_cc)
        elif self.linkage_type == "slider-crank":
            self.draw_slider_crank(
                canvas, phi_drv, mouse_xy, pixels_per_unit, origin_cc)

    def draw_crank_slider(self, canvas, phi_drv, pixels_per_unit,
                          origin_cc):
        self.new_coupler_crank_slider(phi_drv)
        # Get coupler rotation
        delta_theta = (direction(self.s_fol - self.k_drv)
                       - self.phi_cpl_0)
        rotation = rot(delta_theta)
        # Get new position axis (center and angle)
        axis_ctr = self.k_drv + rotation * self.to_axis_ctr
        theta = self.theta_0 + delta_theta
        # Get new coupler border (center and top-right vector)
        border_ctr = axis_ctr + rot(theta) * self.border_ctr_0
        # Create figures to draw
        figures = []
        figures.append(
            CirclePoint(self.k_drv, "Anim", ANIMATION_COLOR))
        figures.append(Link(self.m_drv, self.k_drv, pixels_per_unit,
                            "Anim", ANIMATION_COLOR))
        figures.append(
            SliderPoint(self.s_fol, self.s_fol_angle, "Anim",
                        ANIMATION_COLOR))
        figures.append(
            Axis(axis_ctr, theta, "Anim", ANIMATION_COLOR))
        self.append_sketches(figures, axis_ctr, theta)
        if self.design.show_frame:
            figures.append(CouplerFrame(border_ctr, self.border_tr_0,
                                        theta, pixels_per_unit,
                                        "Anim", ANIMATION_COLOR))
        for the_figure in figures:
            the_figure.render(canvas, the_figure.center_uc,
                              the_figure.angle, pixels_per_unit,
                              origin_cc, ANIMATION_COLOR)

    def draw_double_crank(self, canvas, phi_drv, pixels_per_unit,
                          origin_cc):
        self.new_coupler_double_crank(phi_drv)
        # Get coupler rotation
        delta_theta = (direction(self.k_fol - self.k_drv)
                       - self.phi_cpl_0)
        rotation = rot(delta_theta)
        # Get new position axis (center and angle)
        axis_ctr = self.k_drv + rotation * self.to_axis_ctr
        theta = self.theta_0 + delta_theta
        # Get new coupler border (center and top-right vector)
        border_ctr = axis_ctr + rot(theta) * self.border_ctr_0
        # Create figures to draw
        figures = []
        figures.append(
            CirclePoint(self.k_drv, "Anim", ANIMATION_COLOR))
        figures.append(Link(self.m_drv, self.k_drv, pixels_per_unit,
                            "Anim", ANIMATION_COLOR))
        figures.append(
            CirclePoint(self.k_fol, "Anim", ANIMATION_COLOR))
        figures.append(Link(self.m_fol, self.k_fol, pixels_per_unit,
                            "Anim", ANIMATION_COLOR))
        figures.append(
            Axis(axis_ctr, theta, "Anim", ANIMATION_COLOR))
        self.append_sketches(figures, axis_ctr, theta)
        if self.design.show_frame:
            figures.append(CouplerFrame(border_ctr, self.border_tr_0,
                                        theta, pixels_per_unit,
                                        "Anim", ANIMATION_COLOR))
        for the_figure in figures:
            the_figure.render(canvas, the_figure.center_uc,
                              the_figure.angle, pixels_per_unit,
                              origin_cc, ANIMATION_COLOR)

    def draw_double_slider(self, canvas, phi_drv, mouse_xy,
                           pixels_per_unit, origin_cc):
        self.new_coupler_double_slider(phi_drv, mouse_xy)
        # Get coupler rotation
        delta_theta = (direction(self.s_fol - self.s_drv)
                       - self.phi_cpl_0)
        rotation = rot(delta_theta)
        # Get new position axis (center and angle)
        axis_ctr = self.s_drv + rotation * self.to_axis_ctr
        theta = self.theta_0 + delta_theta
        # Get new coupler border (center and top-right vector)
        border_ctr = axis_ctr + rot(theta) * self.border_ctr_0
        # Create figures to draw
        figures = []
        figures.append(
            SliderPoint(self.s_drv, phi_drv, "Anim", ANIMATION_COLOR))
        figures.append(
            SliderPoint(self.s_fol, self.s_fol_angle, "Anim",
                        ANIMATION_COLOR))
        figures.append(
            Axis(axis_ctr, theta, "Anim", ANIMATION_COLOR))
        self.append_sketches(figures, axis_ctr, theta)
        if self.design.show_frame:
            figures.append(CouplerFrame(border_ctr, self.border_tr_0,
                                        theta, pixels_per_unit,
                                        "Anim", ANIMATION_COLOR))
        for the_figure in figures:
            the_figure.render(canvas, the_figure.center_uc,
                              the_figure.angle, pixels_per_unit,
                              origin_cc, ANIMATION_COLOR)

    def draw_slider_crank(self, canvas, phi_drv, mouse_xy,
                          pixels_per_unit, origin_cc):
        self.new_coupler_slider_crank(phi_drv, mouse_xy)
        # Get coupler rotation
        delta_theta = (direction(self.k_fol - self.s_drv)
                       - self.phi_cpl_0)
        rotation = rot(delta_theta)
        # Get new position axis (center and angle)
        axis_ctr = self.s_drv + rotation * self.to_axis_ctr
        theta = self.theta_0 + delta_theta
        # Get new coupler border (center and top-right vector)
        border_ctr = axis_ctr + rot(theta) * self.border_ctr_0
        # Create figures to draw
        figures = []
        figures.append(
            SliderPoint(self.s_drv, phi_drv, "Anim", ANIMATION_COLOR))
        figures.append(
            CirclePoint(self.k_fol, "Anim", ANIMATION_COLOR))
        figures.append(Link(self.m_fol, self.k_fol, pixels_per_unit,
                            "Anim", ANIMATION_COLOR))
        figures.append(
            Axis(axis_ctr, theta, "Anim", ANIMATION_COLOR))
        self.append_sketches(figures, axis_ctr, theta)
        if self.design.show_frame:
            figures.append(CouplerFrame(border_ctr, self.border_tr_0,
                                        theta, pixels_per_unit,
                                        "Anim", ANIMATION_COLOR))
        for the_figure in figures:
            the_figure.render(canvas, the_figure.center_uc,
                              the_figure.angle, pixels_per_unit,
                              origin_cc, ANIMATION_COLOR)

    def erase(self, canvas):
        canvas.delete("Anim")

    def new_coupler_crank_slider(self, phi_drv):
        """Calculate circlepoint and sliderpoint in new position.
        Leaves them unchanged if position is impossible.
        """
        k_drv_new = self.m_drv + self.drv_length * rot(phi_drv)
        # Calculate s_fol using line-circle intersection
        p1 = self.s_fol_0
        p2 = p1 + self.cpl_length * rot(self.s_fol_angle)
        intsec = intersection_line_circle(p1, p2, k_drv_new,
                                          self.cpl_length)
        if intsec[0] == 2:
            # Two intersections; choose s_fol closest to previous
            if (abs(intsec[1] - self.s_fol)
                < abs(intsec[2] - self.s_fol)):
                self.s_fol = intsec[1]
            else:
                self.s_fol = intsec[2]
            self.k_drv = k_drv_new
        elif intsec[0] == 1:
            # Unique intersection
            self.s_fol = intsec[1]
            self.k_drv = k_drv_new
        else:
            # No intersection, past limits; keep k_drv and s_fol
            pass
        
    def new_coupler_double_crank(self, phi_drv):
        """Calculate circlepoints in new position.
        Leaves them unchanged if position is impossible.
        """
        k_drv_new = self.m_drv + self.drv_length * rot(phi_drv)
        # Calculate k_fol using circle intersection
        intsec = intersection_2_circles(k_drv_new, self.cpl_length,
                                        self.m_fol, self.fol_length)
        if intsec[0] == 2:
            # Two intersections; choose k_fol closest to previous
            if (abs(intsec[1] - self.k_fol)
                < abs(intsec[2] - self.k_fol)):
                self.k_fol = intsec[1]
            else:
                self.k_fol = intsec[2]
            self.k_drv = k_drv_new
        elif intsec[0] == 1:
            # Unique intersection
            self.k_fol = intsec[1]
            self.k_drv = k_drv_new
        else:
            # No intersection, past limits; keep k_drv and k_fol
            pass
            # When displaying angle, reset to arg(k_drv-m_drv)

    def new_coupler_double_slider(self, phi_drv, mouse_xy):
        """Calculate sliderpoints in new position.
        Leaves them unchanged if position is impossible.
        """
        s_path = ComplexLine(self.s_drv_0, self.cpl_length *
                             rot(phi_drv))
        s_drv_new = projection(mouse_xy, s_path)
        # Calculate s_fol using line-circle intersection
        p1 = self.s_fol_0
        p2 = p1 + self.cpl_length * rot(self.s_fol_angle)
        intsec = intersection_line_circle(p1, p2, s_drv_new,
                                          self.cpl_length)
        if intsec[0] == 2:
            # Two intersections; choose s_fol closest to previous
            if (abs(intsec[1] - self.s_fol)
                < abs(intsec[2] - self.s_fol)):
                self.s_fol = intsec[1]
            else:
                self.s_fol = intsec[2]
            self.s_drv = s_drv_new
        elif intsec[0] == 1:
            # Unique intersection
            self.s_fol = intsec[1]
            self.s_drv = s_drv_new
        else:
            # No intersection, past limits; keep s_drv and s_fol
            pass
 
    def new_coupler_slider_crank(self, phi_drv, mouse_xy):
        """Calculate sliderpoint and circlepoint in new position.
        Leaves them unchanged if position is impossible.
        """
        s_path = ComplexLine(self.s_drv_0, self.cpl_length *
                             rot(phi_drv))
        s_drv_new = projection(mouse_xy, s_path)
        # Calculate k_fol using circle intersection
        intsec = intersection_2_circles(s_drv_new, self.cpl_length,
                                        self.m_fol, self.fol_length)
        if intsec[0] == 2:
            # Two intersections; choose k_fol closest to previous
            if (abs(intsec[1] - self.k_fol)
                < abs(intsec[2] - self.k_fol)):
                self.k_fol = intsec[1]
            else:
                self.k_fol = intsec[2]
            self.s_drv = s_drv_new
        elif intsec[0] == 1:
            # Unique intersection
            self.k_fol = intsec[1]
            self.s_drv = s_drv_new
        else:
            # No intersection, past limits; keep s_drv and k_fol
            pass
      
    def redraw(self, design, canvas, phi_drv, mouse_xy,
               pixels_per_unit, origin_cc):
        self.erase(canvas)
        design.update_all(canvas)
        self.draw(canvas, phi_drv, mouse_xy, pixels_per_unit,
                  origin_cc)
