#!/usr/bin/env python3
"""main module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

from tkinter import *
from tkinter import filedialog
from tkinter.filedialog import askopenfilename, asksaveasfilename
import datetime
import json
import webbrowser
from pathlib import Path

# Indirect imports:
# from global_constants import *
# from graphics import *
# from kinematic_symbols import *
# from design import *
# from geometry import *

from button_events import *
from mouse_events import *

class StatusBar(Frame):

    def __init__(self, master):
        Frame.__init__(self, master)
        self.label = Label(self, bd=1, relief=SUNKEN, anchor=W,
                           background=STATUS_DEFAULT_COLOR)
        self.label.pack(fill=X)

    def clear(self):
        self.label.config(text="", background=STATUS_DEFAULT_COLOR)
        self.label.update_idletasks()

    def set(self, format, *args):
        self.clear()
        self.label.config(text=format % args)
        self.label.update_idletasks()

    def update(self, **args):
        self.label.config(args)
        self.label.update_idletasks()

class App(Frame):

    def __init__(self, master):
        # Create a desktop to display the design and provide tools
        Frame.__init__(self, master)
        Pack.config(self, fill=BOTH, expand=TRUE)
        screenwidth = self.winfo_screenwidth()
        screenheight = self.winfo_screenheight()
        canvas_width_px = 0.75 * screenwidth
        canvas_height_px = 0.75 * screenheight
        self.default_origin_cc = complex(50, canvas_height_px-50)
        self.last_file_path = ""
        self.filetypes = (('JSON files', '*.json'),
                          ('All files', '*.*'))
        self.setting_origin = False
        self.task = "Design"
        self.sketch_mode = None
        self.create_widgets(canvas_width_px, canvas_height_px)
        self.canvas.configure(cursor=DEFAULT_CURSOR)
        self.splash(canvas_width_px, canvas_height_px)
        self.new_design()

    def about_ewok(self):
        about_popup = Toplevel(root)
        about_popup.geometry(self.popup_geometry())
        about_popup.title("About EWOK")
        about_ewok = """\
        Copyright © 2022 Theodore B. Ruegsegger

        EWOK is an interactive graphic program for synthesizing
        four-bar linkages to produce specific coupler motions.

        EWOK is a reincarnation of a program I wrote in 1974, part of
        the KinSyn ("Kinematic Synthesis") series under the guidance
        of my thesis advisor and mentor Prof. Roger E. Kaufman. It
        replaces some truly horrid FORTRAN code with Python+Tkinter
        and what I hope counts as good modern practice. "EWOK" is a
        recursive acronym for "EWOK Was Once KinSyn."
        """
        about_label = \
            Label(about_popup, text=about_ewok).pack(side=LEFT)
        close_about_button = Button(about_popup, text="Close",
                                    background=BG_HIGHLIGHT_COLOR,
                                    command=about_popup.destroy)
        close_about_button.pack(side=BOTTOM)

    def add_sketch(self):
        self.sketch_mode = "Add Sketch"
        self.canvas.configure(cursor=EDIT_CURSOR)
    
    def ask_new_origin(self):
        """Set new origin where user clicks left"""
        prompt = "Move mouse to desired origin and left-click."
        self.canvas.statusbar.update(text=prompt,
                                     background=BG_HIGHLIGHT_COLOR)
        # Divert left-click over canvas with global flag so it passes
        # coordinates to function set_origin:
        self.setting_origin = True

    def ask_new_scale(self):
        """
        Display pixels_per_unit and unit in entry fields for user to
        change, wrapped in a popup box a third from top of window
        """
        def update_scale():
            ppu = ppu_str.get()
            new_unit = uu_str.get()
            old_ppu = self.design.pixels_per_unit
            try:
                new_ppu = int(ppu)
            except ValueError:
                prompt = "Error: pixels must be an integer."
                prompt += " Try again."
                self.canvas.statusbar.update(
                    text=prompt, background=BG_HIGHLIGHT_COLOR)
                enter_ppu.delete(0, END)
                enter_ppu.insert(END, str(int(old_ppu)))
            else:
                self.design.unit = new_unit
                self.design.pixels_per_unit = new_ppu
                self.canvas.addtag_all("ToDelete")
                self.canvas.delete("ToDelete")
                if self.task == "Animate" or self.task == "Design":
                    self.canvas.fixed_axes.redraw(
                        self.canvas, self.design.origin_cc, new_unit,
                        new_ppu, FIXED_PLANE_COLOR)
                    # Redraw everything else
                    self.design.update_all(self.canvas)
                elif self.task == "Sketch Coupler":
                    self.canvas.fixed_axes.redraw(
                        self.canvas, self.design.origin_cc, new_unit,
                        new_ppu, DESIGN_POSITION_COLOR)
                    self.ref_axis.redraw(self.canvas, 0+0j, 0,
                                         new_ppu,
                                         self.design.origin_cc,
                                         HIGHLIGHT_COLOR)
                    sketch_resume(self.design, self.canvas, self.task)
                elif self.task == "Sketch Fixed":
                    self.canvas.fixed_axes.redraw(
                        self.canvas, self.design.origin_cc, new_unit,
                        new_ppu, FIXED_PLANE_COLOR)
                    sketch_resume(self.design, self.canvas, self.task)
                self.canvas.statusbar.clear()
                scale_popup.destroy()
        scale_popup = Toplevel(root)
        scale_popup.geometry(self.popup_geometry())
        scale_popup.title("Update Scale and Unit")
        scale = Label(scale_popup, text = "Scale: ").pack(side=LEFT)
        ppu_str = StringVar()
        enter_ppu = Entry(scale_popup, textvariable=ppu_str, width=6,
                          justify=RIGHT)
        enter_ppu.insert(END, str(int(self.design.pixels_per_unit)))
        enter_ppu.pack(side=LEFT)
        pixels_per = Label(scale_popup,
                           text = "pixels per ").pack(side=LEFT)
        uu_str = StringVar()
        enter_uu = Entry(scale_popup, textvariable=uu_str, width=5)
        enter_uu.insert(END, self.design.unit)
        enter_uu.pack(side=LEFT)
        update_scale_button = Button(scale_popup, text = "Update",
                                     background=BG_HIGHLIGHT_COLOR,
                                     command=update_scale)
        update_scale_button.pack(side=LEFT)
        
    def click_left(self, event):
        if self.setting_origin:
            self.set_origin(event)
            self.design.mode = "Select"
        elif self.task == "Animate":
            animate_click(self.design, self.canvas, event)
        elif self.task == "Design":
            left_click(self.design, self.canvas, event)
        elif self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_click(self.design, self.canvas, self.sketch_mode,
                         self.task, event)
            if self.sketch_mode == "Delete Sketch":
                self.sketch_mode = None
                self.canvas.configure(cursor=DEFAULT_CURSOR)

    def click_left_double(self, event):
        if self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_click_double(self.design, self.canvas,
                                self.sketch_mode, self.task, event)
            self.sketch_mode = None

    def click_middle(self, event):
        middle_click(self.design, self.canvas, event)

    def click_right(self, event):
        if self.task == "Design":
            right_click(self.design, self.canvas, event)
        elif self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_click_right(self.design, self.canvas,
                               self.sketch_mode, event)

    def create_toolbar(self):
        toolbar = Frame(self)
        toolbar.pack(side=TOP, fill=X)
        position_button = Button(toolbar, height=30,
                                 background=DESIGN_POSITION_COLOR,
                                 foreground='white',
                                 bitmap="@images/axis.xbm",
                                 activeforeground=HIGHLIGHT_COLOR,
                                 command=self.new_position)
        position_button.pack(side=LEFT)
        circlepoint_button = Button(toolbar,
                                    height=30,
                                    background=DESIGN_POSITION_COLOR, 
                                    foreground='white', 
                                    bitmap="@images/circlept.xbm",
                                    activeforeground=HIGHLIGHT_COLOR,
                                    command=self.new_circlepoint)
        circlepoint_button.pack(side=LEFT)
        sliderpoint_button = Button(toolbar,
                                    height=30,
                                    background=DESIGN_POSITION_COLOR, 
                                    foreground='white', 
                                    bitmap="@images/sliderpt.xbm",
                                    activeforeground=HIGHLIGHT_COLOR,
                                    command=self.new_sliderpoint)
        sliderpoint_button.pack(side=LEFT)
        centerpoint_button = Button(toolbar,
                                    height=30,
                                    background=FIXED_PLANE_COLOR, 
                                    foreground='white',
                                    bitmap="@images/centerpt.xbm",
                                    activeforeground=HIGHLIGHT_COLOR,
                                    command=self.new_centerpoint)
        centerpoint_button.pack(side=LEFT)
        turning_block_button = Button(
            toolbar, height=30, background=FIXED_PLANE_COLOR,
            foreground='white', bitmap="@images/turnblk.xbm",
            activeforeground=HIGHLIGHT_COLOR,
            command=self.new_turning_block)
        turning_block_button.pack(side=LEFT)
        add_sketch_button = Button(
            toolbar, height=30, background=SKETCH_COUPLER_TASK_COLOR,
            foreground='white', bitmap="@images/sketch.xbm",
            activeforeground=HIGHLIGHT_COLOR, command=self.add_sketch)
        add_sketch_button.pack(side=LEFT)
        delete_object_button = Button(
            toolbar, height=30, background=HIGHLIGHT_COLOR,
            bitmap="@images/pirate.xbm",
            activeforeground=HIGHLIGHT_COLOR,
            command=self.destroy_object)
        delete_object_button.pack(side=LEFT)
        self.design_buttons = [position_button, circlepoint_button,
                               sliderpoint_button, centerpoint_button,
                               turning_block_button,
                               delete_object_button]
        self.sketch_buttons = [add_sketch_button,
                               delete_object_button]
        self.color_sketch_buttons = [add_sketch_button]

    def create_top_level_menu(self):
        def not_yet():
            self.canvas.statusbar.set(
                "%s", "This feature available Real Soon Now!")
        self.menubar = Menu(self, background=DESIGN_TASK_COLOR)
        root.config(menu=self.menubar)
        #
        filemenu = Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label="File", menu=filemenu)
        filemenu.add_command(label="New Design",
                             command=self.new_design)
        filemenu.add_command(label="Open Design",
                             command=self.open_design)
        filemenu.add_command(label="Save Design",
                             command=self.save_design)
        filemenu.add_command(label="Describe Design",
                             command=self.describe_design_ask_title)
        filemenu.add_separator()
        filemenu.add_command(label="Exit",
                             command=root.quit)
        #
        self.editmenu = Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label="Edit", menu=self.editmenu)
        self.editmenu.add_command(label="Undo",
                                  command=not_yet)
        self.editmenu.add_command(label="Add Position",
                                  command=self.new_position)
        self.editmenu.add_command(label="Add Circlepoint",
                                  command=self.new_circlepoint)
        self.editmenu.add_command(label="Add Sliderpoint",
                                  command=self.new_sliderpoint)
        self.editmenu.add_command(label="Add Centerpoint",
                                  command=self.new_centerpoint)
        self.editmenu.add_command(label="Add Turning Block",
                                  command=self.new_turning_block)
        self.editmenu.add_command(label="Delete Object",
                                  command=self.destroy_object)
        self.editmenu.add_command(label="Add Sketch",
                                  command=self.add_sketch)
        self.editmenu.add_command(label="Delete Sketch",
                                  command=self.destroy_object)
        self.design_commands = [1, 2, 3, 4, 5, 6]
        self.sketch_commands = [7, 8]
        #
        taskmenu = Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label="Task", menu=taskmenu)
        self.active_task = StringVar(self, value="Design")
        taskmenu.add_radiobutton(label="Design",
                                 variable=self.active_task,
                                 value="Design",
                                 background=DESIGN_TASK_COLOR,
                                 activebackground=DESIGN_TASK_COLOR,
                                 command=self.set_task_design)
        taskmenu.add_radiobutton(label="Animate",
                                 variable=self.active_task,
                                 value="Animate",
                                 background=ANIMATE_TASK_COLOR,
                                 activebackground=ANIMATE_TASK_COLOR,
                                 command=self.set_task_animate)
        taskmenu.add_radiobutton(
            label="Sketch Coupler", variable=self.active_task,
            value="Sketch_Coupler",
            background=SKETCH_COUPLER_TASK_COLOR,
            activebackground=SKETCH_COUPLER_TASK_COLOR,
            command=self.set_task_sketch_coupler)
        taskmenu.add_radiobutton(
            label="Sketch Fixed Frame", variable=self.active_task,
            value="Sketch_Fixed_Frame",
            background=SKETCH_FIXED_TASK_COLOR,
            activebackground=SKETCH_FIXED_TASK_COLOR,
            command=self.set_task_sketch_fixed)
        # taskmenu.add_radiobutton(label="Analyze",
        #                          variable=self.active_task,
        #                          value="Analyze",
        #                          command=self.set_task_analyze)
        #
        viewmenu = Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label="View", menu=viewmenu)
        self.rotate_checked = BooleanVar(self, value=False)
        viewmenu.add_checkbutton(label="Rotate in 5° Steps",
                                 variable=self.rotate_checked,
                                 onvalue=True, offvalue=False,
                                 command=self.toggle_rotate_by_steps)
        self.frame_checked = BooleanVar(self, value=True)
        viewmenu.add_checkbutton(label="Frame Coupler Points",
                                 variable=self.frame_checked,
                                 onvalue=True, offvalue=False,
                                 command=self.toggle_coupler_frame)
        viewmenu.add_separator()
        viewmenu.add_command(label="Set Diagram Origin",
                             command=self.ask_new_origin)
        viewmenu.add_command(label="Set Diagram Scale",
                             command=self.ask_new_scale)
        #
        help_menu = Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label="Help", menu=help_menu)
        help_menu.add_command(label="About EWOK",
                              command=self.about_ewok)
        help_menu.add_command(label="Documentation",
                              command=self.offer_documents)
        
    def create_widgets(self, canvas_width_px, canvas_height_px):
        self.create_top_level_menu()
        self.create_toolbar()
        self.set_up_canvas(canvas_width_px, canvas_height_px)

    def describe_design(self, title):
        # Construct html doc as a string, then write
        e, f = self.design.extract_design(self.canvas)
        # w is a string to accumulate html output
        w = "<html><head>\n"
        w += f"<title>{title}</title>\n"
        w += "<style>table {border: thin solid}</style>\n"
        w += "<style>td {border-left: thin solid;"
        w += " vertical-align: top}</style>\n"
        w += "<style>.bot {border-bottom: thin solid}</style>\n"
        w += "</head><body>\n"
        w += f"<h1 style=\"text-align: center\">{title}</h1>\n"
        w += f"<p>{datetime.datetime.now().strftime('%d %b %Y, %X')}"
        w += "</p>\n"
        w += f"<p>Scale: {f['pixels_per_unit']} pixels/{f['unit']}"
        w += "</p>\n"
        w += "<table>\n<tbody>\n<tr>\n"
        w += "<th class=\"bot\">Dimensions</th>"
        w += f"<th class=\"bot\">{f['unit']}</th>"
        w += "<th class=\"bot\">pixels</th>"
        w += "</tr>\n"
        w += "<tr><td>Origin</td><td>(0,0)</td><td>("
        w += f"{f['origin_cc_x']},{f['origin_cc_y']})</td></tr>\n"
        w += f"<tr><td>Canvas</td><td>{f['width_uu']}x"
        w += f"{f['height_uu']}</td><td>"
        w += f"{f['canvas_width']}x{f['canvas_height']}</td></tr>\n"
        w += "</tbody>\n</table>\n"
        w += f"<p>Rotate in 5&deg; Steps: {f['rotate_by_steps']}"
        w += "<br />\n"
        w += f"Frame Coupler Points: {f['show_frame']}</p>\n"
        if self.design.is_design_complete():
            w += "<p>Design is complete.</p>\n"
        else:
            w += "<p>Design is incomplete.</p>\n"
        if f['positions_count'] == 0:
            w += "<h2>Linkage synthesis, no precision positions yet"
            w += "</h2>\n"
        else:
            w += self.describe_design_table_headers(f)
        w += self.describe_design_list(
            'positions', 'Position', f)
        w += self.describe_design_list(
            'centerpoints', 'Centerpoint', f)
        w += self.describe_design_list(
            'circlepoints', 'Circlepoint', f)
        w += self.describe_design_list(
            'sliderpoints', 'Sliderpoint', f)
        w += self.describe_design_list(
            'turning_blocks', 'Turning Block', f)
        w += "</tbody>\n</table>\n"
        sketch_count = f['coupler_sketches_count'] \
            + f['fixed_sketches_count']
        if sketch_count > 0:
            if sketch_count == 1:
                name = "sketch"
            else:
                name = "sketches"
            w += f"<h2>Annotated with {sketch_count} {name}:</h2>\n"
            w += "<table>\n<tbody>\n"
            w += "<tr><th class=\"bot\">&nbsp;</th>"
            w += "<th class=\"bot\">Points</th>"
            w += "<th class=\"bot\">Tag</th>"
            w += "<th class=\"bot\">Color</th></tr>\n"
            w += self.describe_design_sketches(
                'coupler_sketches', 'Coupler Sketch', f)
            w += self.describe_design_sketches(
                'fixed_sketches', 'Fixed Sketch', f)
            w += "</tbody>\n</table>\n"
        w += "</body>\n</html>\n"
        htmlfiles = (('HTML files', '*.html'),
                     ('HTML files', '*.htm'),
                     ('All files', '*.*'))
        file_path = asksaveasfilename(title='Save as...',
                                      filetypes=htmlfiles,
                                      defaultextension='.html')
        if file_path == () or file_path == "":
            # User canceled action
            pass
        else:
            with open(file_path, "w") as file:
                file.write(w)
            self.describe_design_offer_link(title, file_path)

    def describe_design_ask_title(self):
        def update_title():
            title = title_str.get()
            title_popup.destroy()
            self.describe_design(title)
        #
        filename = self.filename_from_path(self.last_file_path)
        title = "EWOK"
        if filename != "":
            title += f": {filename}"
        title_popup = Toplevel(root)
        title_popup.geometry(self.popup_geometry())
        title_popup.title("Enter a descriptive title")
        title_str = StringVar()
        enter_title = Entry(title_popup, textvariable=title_str,
                            width=30)
        enter_title.insert(END, title)
        enter_title.pack(side=LEFT)
        update_title_button = Button(title_popup, text = "Continue",
                                     background=BG_HIGHLIGHT_COLOR,
                                     command=update_title)
        update_title_button.pack(side=LEFT)

    def describe_design_list(self, label, singular, f):
        w = ""
        list_len = f[label+"_count"]
        for i in range(list_len):
            rec = f[label][i]
            rounded_points = rec
            x = rec['x']
            y = rec['y']
            if label == 'centerpoints' or label == 'circlepoints':
                angle4 = degrees = "&nbsp;"
            else:
                angle4 = rec['angle']
                degrees = rec['degrees']
            tag = rec['tag']
            color = rec['color']
            w += f"<tr><td>{singular} {i}</td><td>({x},{y})</td>"
            w += f"<td>{angle4}</td><td>{degrees}</td><td>{tag}</td>"
            w += f"<td>{color}</td></tr>\n"
        return w

    def describe_design_offer_link(self, title, file_path):
        def callback(url):
            webbrowser.open_new_tab(url)
        webpage_popup = Toplevel(root)
        webpage_popup.geometry("400x100" + self.popup_geometry())
        webpage_popup.title("Click link to see design description")
        if Path(file_path).exists():
            local_link = Label(webpage_popup, text=title, fg="blue",
                               cursor="hand2")
            local_link.pack()
            local_link.bind("<Button-1>",
                            lambda e: callback(file_path))
        close_button = Button(webpage_popup, text="Close",
                              background=BG_HIGHLIGHT_COLOR,
                              command=webpage_popup.destroy)
        close_button.pack(side=BOTTOM)

    def describe_design_sketches(self, label, singular, f):
        w = ""
        list_len = f[label+"_count"]
        for i in range(list_len):
            rec = f[label][i]
            points = rec['points']
            tag = rec['tag']
            color = rec['color']
            w += f"<tr><td>{singular} {i}</td><td>{points}</td>"
            w += f"<td>{tag}</td><td>{color}</td></tr>\n"
        return w

    def describe_design_table_headers(self,f):     
        w = f"<h2>Linkage synthesis for {f['positions_count']}"
        w += " precision positions:</h2>\n"
        w += "<table>\n<tbody>\n"
        w += "<tr><th>&nbsp;</th>"
        w += "<th style=\"border-left: none\">Center</th>"
        w += "<th colspan=\"2\">Angle</th><th>Tag</th><th>Color</th>"
        w += "</tr>\n"
        w += "<tr><th class=\"bot\" colspan=\"2\">&nbsp;</th>"
        w += "<th class=\"bot\">Radians</th>"
        w += "<th class=\"bot\">Degrees</th>"
        w += "<th class=\"bot\">&nbsp;</th>"
        w += "<th class=\"bot\">&nbsp;</th>"
        w += "</tr>\n"
        return w
        
    def destroy_object(self):
        if self.task == "Design":
            delete_object(self.design, self.canvas)
        elif self.task == "Sketch Coupler":
            self.destroy_sketch(self.design.coupler_sketches)
        elif self.task == "Sketch Fixed":
            self.destroy_sketch(self.design.fixed_sketches)

    def destroy_sketch(self, sketch_list):
        # User wants to delete a sketch. Sets mode and cursor
        if len(sketch_list) > 0:
            self.sketch_mode = "Delete Sketch"
            self.canvas.configure(cursor=DELETE_CURSOR)
        else:
            self.canvas.configure(cursor=DEFAULT_CURSOR)
            self.sketch_mode = None

    def disable_design_buttons(self):
        for button in self.design_buttons:
            button.configure(state=DISABLED)
    
    def disable_design_menu(self):
        for index in self.design_commands:
            self.editmenu.entryconfigure(index, state=DISABLED)

    def disable_sketch_buttons(self):
        for button in self.sketch_buttons:
            button.configure(state=DISABLED)
        
    def disable_sketch_menu(self):
        for index in self.sketch_commands:
            self.editmenu.entryconfigure(index, state=DISABLED)

    def enable_design_buttons(self):
        for button in self.design_buttons:
            button.configure(state=NORMAL)

    def enable_design_menu(self):
        for index in self.design_commands:
            self.editmenu.entryconfigure(index, state=NORMAL)

    def enable_sketch_buttons(self):
        for button in self.sketch_buttons:
            button.configure(state=NORMAL)
            
    def enable_sketch_menu(self):
        for index in self.sketch_commands:
            self.editmenu.entryconfigure(index, state=NORMAL)
            
    def filename_from_path(self, path):
        # strip any directories and extensions
        return path[path.rfind("/")+1:path.rfind(".")]

    def mouse_leave_canvas(self, event):
        leave_canvas(self.design, self.canvas, self.task, event)

    def move_left(self, event):
        if self.task == "Animate":
            animate_move(self.design, self.canvas, event)
        elif self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_move(self.design, self.canvas, self.sketch_mode,
                        event)
        else:
            left_move(self.design, self.canvas, event)
        
    def move_middle(self, event):
        middle_move(self.design, self.canvas, event)

    def move_mouse(self, event):
        # Mouse moves with no buttons clicked
        if self.setting_origin:
            text=f"({event.x},{event.y}) in canvas coordinates"
            text += " (pixels from top left)"
            self.canvas.statusbar.update(text=text)
        elif self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_mouse_motion(self.design, self.canvas,
                                self.sketch_mode, self.task, event)
        elif self.task == "Design" or self.task == "Animate":
            mouse_motion(self.design, self.canvas, event)

    def move_right(self, event):
        if self.task == "Design":
            right_move(self.design, self.canvas, event)
        elif self.task == "Sketch Coupler" \
             or self.task == "Sketch Fixed":
            sketch_move_right(self.design, self.canvas,
                              self.sketch_mode, event)

    def new_centerpoint(self):
        if self.task == "Design":
            add_centerpoint(self.design, self.canvas)

    def new_circlepoint(self):
        if self.task == "Design":
            add_circlepoint(self.design, self.canvas)

    def new_design(self):
        # Create a design to contain all the kinematic objects
        self.design = LinkageDesign(DEFAULT_USER_UNIT,
                                    DEFAULT_PIXELS_PER_UNIT,
                                    self.default_origin_cc)
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        self.canvas.fixed_axes.redraw(self.canvas,
                                      self.default_origin_cc,
                                      DEFAULT_USER_UNIT,
                                      DEFAULT_PIXELS_PER_UNIT)
        # Reset default checkboxes in viewmenu
        self.rotate_checked.set(False)
        self.design.rotate_by_steps = False
        self.frame_checked.set(True)
        self.design.show_frame = True
        if self.task == "Design":
            self.set_task_design()
        elif self.task == "Animate":
            self.set_task_animate()
        else:
            self.set_task_design()
        root.title("ewok")

    def new_position(self):
        if self.task == "Design":
            add_position(self.design, self.canvas)

    def new_sliderpoint(self):
        if self.task == "Design":
            add_sliderpoint(self.design, self.canvas)

    def new_turning_block(self):
        if self.task == "Design":
            add_turning_block(self.design, self.canvas)

    def offer_documents(self):
        def callback(url):
            webbrowser.open_new_tab(url)
        docs_popup = Toplevel(root)
        docs_popup.geometry("300x200" + self.popup_geometry())
        docs_popup.title("Documentation")
        header1 = Label(docs_popup, text="Local Copies:",
                        font="TkHeadingFont")
        header1.pack()
        local_um = "doc/usermanual/um.html"
        if Path(local_um).exists():
            link_local_um = Label(docs_popup, text="EWOK User Manual",
                                  fg="blue", cursor="hand2")
            link_local_um.pack()
            link_local_um.bind("<Button-1>",
                               lambda e: callback(local_um))
        local_release_notes = "ReleaseNotes"
        link_local_release_notes = Label(docs_popup,
                                         text="Release Notes",
                                         fg="blue", cursor="hand2")
        link_local_release_notes.bind("<Button-1>", lambda e:
                                      callback(local_release_notes))
        link_local_release_notes.pack()
        header2 = Label(docs_popup, text="EWOK Website:",
                        font="TkHeadingFont")
        header2.pack()
        url_base = "http://toyrus.ngrok.io/~ted/ewok/"
        url_um = url_base + "doc/usermanual/um.html"
        link_um = Label(docs_popup, text="EWOK User Manual",
                        fg="blue", cursor="hand2")
        link_um.pack()
        link_um.bind("<Button-1>", lambda e: callback(url_um))
        url_release_notes = url_base + "ReleaseNotes"
        link_release_notes = Label(docs_popup, text="Release Notes",
                                   fg="blue", cursor="hand2")
        link_release_notes.bind("<Button-1>",
                                lambda e: callback(url_release_notes))
        link_release_notes.pack()
        close_docs_button = Button(docs_popup, text="Close",
                                   background=BG_HIGHLIGHT_COLOR,
                                   command=docs_popup.destroy)
        close_docs_button.pack(side=BOTTOM)

    def open_design(self):
        file_path = askopenfilename(title='Open File',
                                    filetypes=self.filetypes,
                                    defaultextension='.json')
        if file_path == () or file_path == "":
            # User canceled action
            pass
        else:
            with open(file_path, "r") as file:
                # Try to load, allow user to cancel:
                e = json.load(file)
                # Design selected, clear current
                self.new_design()
                self.design.restore_from_dictionary(self.canvas, e)
                self.last_file_path = file_path
                self.canvas.fixed_axes.redraw(
                    self.canvas, self.design.origin_cc,
                    self.design.unit, self.design.pixels_per_unit)
                self.design.update_all(self.canvas)
                root.title("ewok: "+self.filename_from_path(
                    file_path))
                self.rotate_checked.set(self.design.rotate_by_steps)
                self.frame_checked.set(self.design.show_frame)
                    
    def popup_geometry(self):
        # Return geometry for a small popup box
        root.update_idletasks()
        app_width = self.winfo_reqwidth()
        app_height = self.winfo_reqheight()
        root_x = self.winfo_rootx()
        root_y = self.winfo_rooty()
        box_x = int(root_x + app_width/2) - 150
        box_y = int(root_y + app_height/3)
        return f"+{box_x}+{box_y}"

    def release_left(self, event):
        if self.task == "Animate":
            animate_release(self.design, self.canvas, event)
        elif self.task == "Design":
            left_release(self.design, self.canvas, event)

    def release_middle(self, event):
        middle_release(self.design, self.canvas, event)

    def release_right(self, event):
        if self.task == "Design":
            right_release(self.design, self.canvas, event)

    def save_design(self):
        if self.last_file_path != "":
            file_path = asksaveasfilename(
                title='Save as...', initialfile=self.last_file_path,
                filetypes=self.filetypes, defaultextension='.json')
        else:
            initialfile = "ewok_design.json"
            file_path = asksaveasfilename(title='Save as...',
                                          initialfile=initialfile,
                                          filetypes=self.filetypes,
                                          defaultextension='.json')
        if file_path == () or file_path == "":
            # User canceled action
            pass
        else:
            e, f = self.design.extract_design(self.canvas)
            with open(file_path, "w") as file:
                json.dump(e, file, indent=4)
                self.last_file_path = file_path
                root.title("ewok: "+self.filename_from_path(
                    file_path))

    def scrollcanvasX(self, *args):
######################################################################
        """ Not sure what this is or why """
######################################################################
        print("scrolling", args)
        print(self.canvas.scrollX.get())

    def set_color_sketch_buttons(self, color):
        for button in self.color_sketch_buttons:
            button.configure(background=color)

    def set_origin(self, event):
        """Get coords of left-click as new origin_cc, then redraw all
        pictures on canvas"""
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        new_origin_cc = complex(event.x, event.y)
        if self.task == "Sketch Coupler":
            color = DESIGN_POSITION_COLOR
        else:
            color = FIXED_PLANE_COLOR
        self.canvas.fixed_axes.redraw(self.canvas, new_origin_cc,
                                      self.design.unit,
                                      self.design.pixels_per_unit,
                                      color)
        self.design.origin_cc = new_origin_cc
        if self.task == "Animate" or self.task == "Design":
            # Redraw everything else
            self.design.update_all(self.canvas)
        elif self.task == "Sketch Coupler":
            self.ref_axis.redraw(self.canvas, 0+0j, 0,
                                 self.design.pixels_per_unit,
                                 new_origin_cc, HIGHLIGHT_COLOR)
            sketch_resume(self.design, self.canvas, self.task)
        elif self.task == "Sketch Fixed":
            sketch_resume(self.design, self.canvas, self.task)
        self.canvas.statusbar.clear()
        self.setting_origin = False

    def set_task_animate(self):
        self.task = "Animate"
        self.active_task.set("Animate")
        self.design.mode = "Select"
        self.canvas.configure(cursor=ANIMATE_CURSOR)
        self.menubar.configure(background=ANIMATE_TASK_COLOR)
        self.disable_design_buttons()
        self.disable_design_menu()
        self.disable_sketch_buttons()
        self.disable_sketch_menu()
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        self.canvas.fixed_axes.redraw(self.canvas,
                                      self.design.origin_cc,
                                      self.design.unit,
                                      self.design.pixels_per_unit)
        self.canvas.statusbar.update(text="Task is \"Animate\"",
                                     background=ANIMATE_TASK_COLOR)
        self.design.update_all(self.canvas)

    def set_task_design(self):
        self.task = "Design"
        self.active_task.set("Design")
        self.canvas.configure(cursor=DEFAULT_CURSOR)
        self.menubar.configure(background=DESIGN_TASK_COLOR)
        self.disable_sketch_buttons()
        self.disable_sketch_menu()
        self.enable_design_buttons()
        self.enable_design_menu()
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        self.canvas.fixed_axes.redraw(self.canvas,
                                      self.design.origin_cc,
                                      self.design.unit,
                                      self.design.pixels_per_unit)
        self.canvas.statusbar.update(text="Task is \"Design\"",
                                     background=DESIGN_TASK_COLOR)
        self.design.update_all(self.canvas)
        self.design.mode = "Select"

    def set_task_sketch_coupler(self):
        self.task = "Sketch Coupler"
        self.active_task.set("Sketch_Coupler")
        self.canvas.configure(cursor=DEFAULT_CURSOR)
        self.menubar.configure(background=SKETCH_COUPLER_TASK_COLOR)
        self.disable_design_buttons()
        self.disable_design_menu()
        self.enable_sketch_buttons()
        self.set_color_sketch_buttons(SKETCH_COUPLER_TASK_COLOR)
        self.enable_sketch_menu()
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        self.canvas.statusbar.update(
            text="Task is \"Sketch Coupler\"",
            background=SKETCH_COUPLER_TASK_COLOR)
        self.canvas.fixed_axes.redraw(self.canvas,
                                      self.design.origin_cc,
                                      self.design.unit,
                                      self.design.pixels_per_unit,
                                      DESIGN_POSITION_COLOR)
        self.ref_axis = Axis(0+0j, 0, "ref_axis", HIGHLIGHT_COLOR)
        self.ref_axis.render(self.canvas, 0+0j, 0,
                             self.design.pixels_per_unit,
                             self.design.origin_cc, HIGHLIGHT_COLOR)
        self.sketch_mode = None
        sketch_resume(self.design, self.canvas, self.task)
         
    def set_task_sketch_fixed(self):
        self.task = "Sketch Fixed"
        self.active_task.set("Sketch_Fixed")
        self.canvas.configure(cursor=DEFAULT_CURSOR)
        self.menubar.configure(background=SKETCH_FIXED_TASK_COLOR)
        self.disable_design_buttons()
        self.disable_design_menu()
        self.enable_sketch_buttons()
        self.set_color_sketch_buttons(SKETCH_FIXED_TASK_COLOR)
        self.enable_sketch_menu()
        self.canvas.addtag_all("ToDelete")
        self.canvas.delete("ToDelete")
        self.canvas.statusbar.update(
            text="Task is \"Sketch Fixed Frame\"",
            background=SKETCH_FIXED_TASK_COLOR)
        self.canvas.fixed_axes.redraw(self.canvas,
                                      self.design.origin_cc,
                                      self.design.unit,
                                      self.design.pixels_per_unit,
                                      FIXED_PLANE_COLOR)
        self.sketch_mode = None
        sketch_resume(self.design, self.canvas, self.task)

    def set_up_canvas(self, canvas_width_px, canvas_height_px):
        self.canvas = Canvas(self, width=canvas_width_px,
                             height=canvas_height_px)
        self.update_idletasks()
        self.canvas.statusbar = StatusBar(self)
        #self.canvas.scrollX = Scrollbar(self, orient=HORIZONTAL)
        #self.canvas.scrollY = Scrollbar(self, orient=VERTICAL)

        # Now tie the three together. This is standard boilerplate
        # text
        #self.canvas['xscrollcommand'] = self.canvas.scrollX.set
        #self.canvas['yscrollcommand'] = self.canvas.scrollY.set
        #self.canvas.scrollX['command'] = self.canvas.xview
        #self.canvas.scrollY['command'] = self.canvas.yview

        # Pack 'em up
        #self.canvas.scrollX.pack(side=BOTTOM, fill=X)
        #self.canvas.scrollY.pack(side=RIGHT, fill=Y)
        self.canvas.statusbar.pack(side=BOTTOM, fill=X)
        self.canvas.pack(side=LEFT, fill=BOTH, expand=TRUE)

        self.canvas.bind("<Motion>", self.move_mouse)
        self.canvas.bind("<ButtonPress-1>", self.click_left)
        self.canvas.bind("<Double-Button-1>", self.click_left_double)
        self.canvas.bind("<B1-Motion>", self.move_left)
        self.canvas.bind("<ButtonRelease-1>", self.release_left)
        self.canvas.bind("<ButtonPress-2>", self.click_middle)
        self.canvas.bind("<B2-Motion>", self.move_middle)
        self.canvas.bind("<ButtonRelease-2>", self.release_middle)
        self.canvas.bind("<ButtonPress-3>", self.click_right)
        self.canvas.bind("<B3-Motion>", self.move_right)
        self.canvas.bind("<ButtonRelease-3>", self.release_right)
        self.canvas.bind("<Leave>", self.mouse_leave_canvas)
        self.canvas.fixed_axes = FixedPlaneAxisPair(
            self.canvas, self.default_origin_cc, DEFAULT_USER_UNIT,
            DEFAULT_PIXELS_PER_UNIT)

    def splash(self, canvas_width_px, canvas_height_px):
        # Using initial canvas dimensions instead of calling
        # self.popup_geometry() to get around tkinter timing issues
        box_x = int(canvas_width_px/2) - 150
        box_y = int(canvas_height_px/2)
        splash_popup = Toplevel(root)
        splash_popup.title("")
        splash_popup.geometry(f"+{box_x}+{box_y}")
        quote = """\
        Take to kinematics. It will repay you.
        It is more fecund than geometry.
        It adds a fourth dimension to space.
              -- Chebyshev to Sylvester, 1873
        """
        splash_label = \
            Label(splash_popup, text=quote).pack(side=LEFT)
        close_splash_button = Button(splash_popup, text="OK",
                                     background=BG_HIGHLIGHT_COLOR,
                                     command=splash_popup.destroy)
        close_splash_button.pack(side=BOTTOM)

    def toggle_coupler_frame(self):
        self.design.show_frame = self.frame_checked.get()
        if self.design.show_frame:
            show_coupler_frame(self.design, self.canvas)
        else:
            hide_coupler_frame(self.design, self.canvas)

    def toggle_rotate_by_steps(self):
        self.design.rotate_by_steps = self.rotate_checked.get()
        
###############################################################

root = Tk()
root.title("ewok")

ewok = App(root)

root.mainloop()


