"""geometry module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

from math import sqrt

class ComplexLine:
    """A line in complex space: vector passes through origin."""
    def __init__(self, origin, vector):
        self.origin = origin
        self.vector = vector

class Cramer2Solution:
    """Solution by Cramer's rule: X, Y , and determinant."""
    def __init__(self, x, y, determinant):
        self.x = x
        self.y = y
        self.determinant = determinant

class IntersectionPt:
    """Line intersection by Cramer's rule: Point and determinant."""
    def __init__(self, point, determinant):
        self.point = point
        self.determinant = determinant

def cramer2(a1, b1, c1, a2, b2, c2):
    """Return Cramer's rule solution to a pair of linear equations.
    Given coefficients a1, a2, b1, b2, c1, and c2 of equations:
    a1 X + b1 Y = c1
    a2 X + b2 Y = c2
    returns solutions X and Y and the determinant as a set.
    """
    determinant = a1 * b2 - a2 * b1
    if determinant != 0:
        X = (c1 * b2 - c2 * b1) / determinant
        Y = (a1 * c2 - a2 * c1) / determinant
    else:
        X = Y = 0.0
    return Cramer2Solution(X, Y, determinant)

def intersection(complex_line_1, complex_line_2):
    """Return intersection point of two complex lines.
    Returns the point and the Cramer determinant as a set.
    """
    org1 = complex_line_1.origin
    org2 = complex_line_2.origin
    v1 = complex_line_1.vector
    v2 = complex_line_2.vector
    t = org2 + v2 - org1
    solution = cramer2(v1.real, v2.real, t.real,
        v1.imag, v2.imag, t.imag)
    return IntersectionPt(org1 + solution.x * v1,
                          solution.determinant)

def intersection_2_circles(p0, r0, p1, r1):
    """Return circle intersection, using P. Bourke's derivation.
    http://paulbourke.net/geometry/circlesphere/
    Return a tuple (Number of Solutions, Point1, Point2)"""
    p0_to_p1 = p1 - p0
    d = abs(p0_to_p1)
    r_sum = r0 + r1
    r_diff = abs(r0 - r1)
    if r_diff < d < r_sum:
        # Two intersection points
        a = (r0**2 - r1**2 + d**2 ) / (d + d)
        h = sqrt(r0**2 - a**2)
        p2 = p0 + p0_to_p1 * a / d
        p2_to_p3 = (0+1j) * p0_to_p1 * h / d
        solution = (2, p2 + p2_to_p3, p2 - p2_to_p3)
    elif d == r_sum or d == r_diff:
        # One intersection point
        p2 = p0 + p0_to_p1 * r0 / d
        solution = (1, p2, p2)
    else:
        # No intersection points
        solution = (0, 0.0, 0.0)
    return solution

def intersection_line_circle(p1, p2, p3, r):
    """Return circle intersection, using P. Bourke's derivation.
    http://paulbourke.net/geometry/circlesphere/
    Line p2-p1 intersects circle with center p3, radius r
    Return a tuple (Number of Solutions, Point1, Point2)"""
    x1 = p1.real
    y1 = p1.imag
    x2 = p2.real
    y2 = p2.imag
    x3 = p3.real
    y3 = p3.imag
    a = (x2 - x1)**2 + (y2 - y1)**2 
    b = 2*((x2 - x1)*(x1 - x3) + (y2 - y1)*(y1 - y3))
    c = x3**2 + y3**2 + x1**2 + y1**2 - 2*(x3*x1 + y3*y1) - r**2
    discriminant = b**2 - 4*a*c
    if discriminant > 0:
        # Two intersection points
        root = sqrt(discriminant)
        u1 = (-b + root)/(a+a)
        u2 = (-b - root)/(a+a)
        solution = (2, p1 + u1*(p2 - p1), p1 + u2*(p2 - p1))
    elif discriminant == 0:
        # One intersection point
        u = -b/(a+a)
        p = p1 + u*(p2 - p1)
        solution = (1, p, p)
    else:
        # no intersection points
        solution = (0, 0.0, 0.0)
    return solution

def perp_bisector(a, b):
    """ Return perpendicular bisector of line segment from a to b."""
    origin = 0.5 * (a + b)
    vector = (0+1j) * (a - b)
    return ComplexLine(origin, vector)

def projection(point, complex_line):
    """ Return projection of point on complex_line."""
    origin = complex_line.origin
    vector = complex_line.vector
    z = point - origin
    if vector.real == 0:
        # Trivial case: vector is vertical
        proj = origin + z.imag*1j
    elif vector.imag == 0:
        # Trivial case: vector is horizontal
        proj = origin + z.real
    else:
        w = (0+1j) * vector
        proj = origin + vector * (
            z.real * w.imag
            - z.imag * w.real
            ) / (
            vector.real * w.imag
            - vector.imag * w.real)
    return proj

