"""graphics module for EWOK.

Copyright (C) 2022 Theodore B. Ruegsegger
See LICENSE file for terms of use.
"""

from math import atan2, cos, sin, sqrt
from global_constants import *

# Definitions:

# Pixel (px):         screen, canvas, and mouse metric
# unit:               user metric (can be inches, meters, angstroms,
#                     whatever)
# canvas Coordinates: measured in pixels from top left (y downward)
# origin_cc:          location of user's origin, in canvas coordinates
# _uu:                user units (relative displacement)
# _uc:                user coordinates (absolute displacement from
#                     origin, in uu, y upward)
# _cu:                canvas units (relative displacement), aka pixels,
#                     y downward
# _cc:                canvas coordinates (absolute displacement from
#                     top left, in cu, y downward)

# Initial Values
setting_origin = False

def canvas_coordinates(z_uc, pixels_per_unit, origin_cc):
    """Return (complex) canvas coordinates, given user coordinates."""
    return cround(canvas_units(z_uc, pixels_per_unit) + origin_cc)

def canvas_units(z_uu, pixels_per_unit):
    """Return (complex) canvas units, given user units."""
    return z_uu.conjugate() * pixels_per_unit

def cround(z, n=0):
    """Return complex value with parts rounded to n digits."""
    return complex(round(z.real, n), round(z.imag, n));

def direction(z):
    """Return the direction (radians) of a complex number."""
    return atan2(z.imag, z.real)

def rot(angle):
    """Return exp(j*angle) = cos(angle) + j * sin(angle)
    
    This defines the rotation of a complex vector by angle.
    """
    return complex(cos(angle), sin(angle))

def rot_cc(angle):
    """Return exp(j*angle) for canvas units."""
    return complex(cos(angle), -sin(angle))

def user_coordinates(z_cc, pixels_per_unit, origin_cc):
    """Return (complex) user coordinates, given canvas coordinates."""
    return user_units(z_cc - origin_cc, pixels_per_unit)

def user_units(z_cu, pixels_per_unit):
    """Return (complex) user units, given canvas units."""
    return z_cu.conjugate() / pixels_per_unit

# Don't need magnitude; abs works fine for complex numbers.
#def magnitude(z):
#    """Return the magnitude of a complex number."""
#    return sqrt(z.real**2 + z.imag**2)

class Bar_cu:
    """Heavy line, as for a link."""

    def __init__(self, other_end_cu):
        self.other_end_cu = other_end_cu

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        center_cc = canvas_coordinates(center_uc, pixels_per_unit,
                                       origin_cc)
        rotation = rot_cc(angle)
        other_end_cc = cround(canvas_units(center_uc, pixels_per_unit)
            + self.other_end_cu * rotation + origin_cc)
        canvas.create_line(
            center_cc.real, center_cc.imag,
            other_end_cc.real, other_end_cc.imag,
            tags=tag, fill=color, width=2)

class Circle_cu:
    """Centered circle defined by a radius."""

    def __init__(self, radius_cu):
        self.radius_cu = radius_cu

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        center_cc = canvas_coordinates(center_uc, pixels_per_unit,
                                       origin_cc)
        canvas.create_oval(center_cc.real-self.radius_cu,
                           center_cc.imag-self.radius_cu,
                           center_cc.real+self.radius_cu,
                           center_cc.imag+self.radius_cu, tags=tag,
                           outline=color)

class Polyline_cu:
    """Polyline defined as a list of vertices in canvas units (pixels,
    y-axis positive downwards)."""

    def __init__(self, points, is_closed=False):
        self.points = points
        self.is_closed = is_closed

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        rotation = rot_cc(angle)
        canvas_points = []
        for point in self.points:
            canvas_points.append(
                cround(canvas_units(center_uc, pixels_per_unit)
                       + point * rotation + origin_cc))
        for i in range(1, len(canvas_points)):
            canvas.create_line(
                canvas_points[i-1].real, canvas_points[i-1].imag,
                canvas_points[i].real, canvas_points[i].imag,
                tags=tag, fill=color)
        if self.is_closed:
            last = len(canvas_points)-1
            canvas.create_line(
                canvas_points[last].real, canvas_points[last].imag,
                canvas_points[0].real, canvas_points[0].imag,
                tags=tag, fill=color)

class Polygon_cu(Polyline_cu):
    """Polygon defined as a list of vertices. Same as a Polyline_cu
    but closed (last point connects to first).

    """
    
    def __init__(self, points):
        self.points = points
        self.is_closed = True
        
class Polyline_uu:
    """Polyline defined as a list of vertices in user units (y-axis
    positive upwards)."""

    def __init__(self, points, is_closed=False):
        self.points = points
        self.is_closed = is_closed

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        rotation = rot(angle)
        canvas_points = []
        for point in self.points:
            canvas_points.append(
                cround(canvas_units(center_uc + point * rotation,
                                    pixels_per_unit) + origin_cc))
        for i in range(1, len(canvas_points)):
            canvas.create_line(
                canvas_points[i-1].real, canvas_points[i-1].imag,
                canvas_points[i].real, canvas_points[i].imag,
                tags=tag, fill=color)
        if self.is_closed:
            last = len(canvas_points)-1
            canvas.create_line(
                canvas_points[last].real, canvas_points[last].imag,
                canvas_points[0].real, canvas_points[0].imag,
                tags=tag, fill=color)

class Polygon_uu(Polyline_uu):
    """Polygon defined as a list of vertices. Same as a Polyline_uu
    but closed (last point connects to first)."""
    
    def __init__(self, points):
        self.points = points
        self.is_closed = True
        
class QuarterCircle_cu:
    """Quarter-circle defined by an offset vector and a radius."""

    def __init__(self, offset_cu, radius_cu):
        self.offset_cu = offset_cu
        self.radius_cu = radius_cu
        # In cu, Y is positive down, thus flip phi:
        phi = - round(direction(offset_cu) * DEGREES_PER_RADIAN)
        if phi > 0:
            if phi < 90:
                self.start = 0
            else:
                self.start = 90
        else:
            if phi < -90:
                self.start = 180
            else:
                self.start = 270

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        center_cc = cround(canvas_units(center_uc, pixels_per_unit) +
                           self.offset_cu * rot_cc(angle) +
                           origin_cc)
        canvas.create_arc(
            center_cc.real-self.radius_cu,
            center_cc.imag-self.radius_cu,
            center_cc.real+self.radius_cu,
            center_cc.imag+self.radius_cu,
            style="arc",
            start=self.start + round(angle * DEGREES_PER_RADIAN),
            extent=90, tags=tag, outline=color, fill=color)

class Rectangle_cu(Polygon_cu):
    """Centered rectangle defined by a vector from center to corner.

    Rendered as a polygon so it can be rotated."""

    def __init__(self, LowerRight):
        self.points = [
            LowerRight, 
            LowerRight.conjugate(),
            -LowerRight, 
            -LowerRight.conjugate() ]
        self.is_closed = True

class SemiCircle_cu:
    """Centered semicircle defined by a radius, convex up.""" 

    def __init__(self, radius_cu):
        self.radius_cu = radius_cu

    def draw(self, canvas, center_uc, angle, pixels_per_unit,
             origin_cc, tag, color):
        center_cc = canvas_coordinates(center_uc, pixels_per_unit,
                                       origin_cc)
        canvas.create_arc(center_cc.real-self.radius_cu,
                          center_cc.imag-self.radius_cu,
                          center_cc.real+self.radius_cu,
                          center_cc.imag+self.radius_cu,
                          style="arc",
                          start=round(angle * DEGREES_PER_RADIAN),
                          extent=181, tags=tag, outline=color,
                          fill=color)

